"""
test all the datasets in the load-atoms database
"""

import shutil
from pathlib import Path

import pytest
from load_atoms import load_dataset
from load_atoms.dataset import DescribedDataset
from load_atoms.utils import matches_checksum

# this file is at root/tests/database/test_all_datasets.py
project_root = Path(__file__).parent.parent

# database is at root/database
databaset_root = project_root / "database"
dataset_names = sorted([p.name for p in databaset_root.iterdir() if p.is_dir()])


# copy over the datasets to a place to test them
testing_dir = project_root / "testing-datasets"
testing_dir.mkdir(exist_ok=True)

for name in dataset_names:
    # copy over the folder if it doesn't exist
    if not (testing_dir / name).exists():
        shutil.copytree(databaset_root / name, testing_dir / name)


@pytest.mark.parametrize("name", dataset_names, ids=dataset_names)
def test_dataset(name):
    """Test that the dataset can be loaded."""

    # ignore the really big datasets to save on time
    to_ignore = ["C-SYNTH-23M"]
    if name in to_ignore:
        pytest.skip(f"Skipping {name}")

    ds: DescribedDataset = load_dataset(name, root=testing_dir)  # type: ignore

    # check that all files match their checksums
    for filename, hash in ds.description.files.items():
        file = testing_dir / name / filename
        assert matches_checksum(file, hash)


@pytest.mark.parametrize("name", dataset_names, ids=dataset_names)
def test_docs_exist(name):
    """Test that documentation exists for each dataset."""

    docs = project_root / "docs/source/datasets"

    assert (docs / f"{name}.rst").exists(), f"Missing docs for {name}"



from __future__ import annotations

from typing import Any, Callable, Sequence

import ase
import numpy as np

from load_atoms.dataset import AtomsDataset

__all__ = ["filter_by", "cross_validate_split"]


def filter_by(
    dataset: AtomsDataset | Sequence[ase.Atoms],
    *functions: Callable[[ase.Atoms], bool],
    **info_kwargs: Any,
) -> AtomsDataset:
    """
    Filter a dataset.

    Parameters
    ----------
    dataset
        The dataset to filter.
    functions
        Functions to filter the dataset by. Each function should take an
        ASE Atoms object as input and return a boolean.
    info_kwargs
        Keyword arguments to filter the dataset by. Only atoms objects with
        matching info keys and values will be returned.

    Returns
    -------
    AtomsDataset
        The filtered dataset.

    Examples
    --------
    >>> from ase import Atoms
    >>> from load_atoms.manipulations import filter_by
    >>> structures = [
    ...     Atoms("H2O", info=dict(name="water")),
    ...     Atoms("H2O2", info=dict(name="hydrogen peroxide")),
    ...     Atoms("CH4", info=dict(name="methane")),
    ... ]
    >>> small = filter_by(
    ...     structures,
    ...     lambda structure: len(structure) < 4,
    ... )
    >>> len(small)
    1
    >>> water = filter_by(name="water")
    >>> len(water)
    1
    """

    def matches_info(structure: ase.Atoms) -> bool:
        for key, value in info_kwargs.items():
            if structure.info.get(key, None) != value:
                return False
        return True

    functions = (*functions, matches_info)

    def the_filter(structure: ase.Atoms) -> bool:
        return all(function(structure) for function in functions)

    return AtomsDataset(
        [structure for structure in dataset if the_filter(structure)]
    )


def cross_validate_split(
    dataset: AtomsDataset | Sequence[ase.Atoms],
    fold,
    k: int = 5,
    n_test: int | None = None,
    seed: int = 0,
) -> tuple[AtomsDataset, AtomsDataset]:
    """
    Generate a shuffled train/test split for cross-validation.

    Parameters
    ----------
    dataset
        The dataset to split.
    fold : int
        The fold to use for testing.
    k
        The number of folds to use, by default 5.
    n_test
        The number of structures to use for testing, by
        this will be set to len(dataset) // k.
    seed
        The random seed to use, by default 0.

    Returns
    -------
    Tuple[Dataset, Dataset]
        The train and test datasets.

    Examples
    --------
    >>> from ase import Atoms
    >>> from load_atoms.manipulations import cross_validate_split
    >>> structures = [
    ...     Atoms("H2O", info=dict(name="water")),
    ...     Atoms("H2O2", info=dict(name="hydrogen peroxide")),
    ...     Atoms("CH4", info=dict(name="methane")),
    ... ]
    >>> train, test = cross_validate_split(structures, fold=0, k=3)
    >>> len(train)
    2
    >>> len(test)
    1
    """

    if n_test is None:
        n_test = len(dataset) // k

    idxs = np.arange(len(dataset))
    np.random.RandomState(seed).shuffle(idxs)
    idxs = np.roll(idxs, fold * len(idxs) // k)
    train, test = idxs[:-n_test], idxs[-n_test:]

    return (
        AtomsDataset([dataset[t] for t in train]),
        AtomsDataset([dataset[t] for t in test]),
    )



from __future__ import annotations

import hashlib
import string
from pathlib import Path
from typing import Callable, Generic, Iterable, Sequence, TypeVar

FRONTEND_URL = "https://jla-gardner.github.io/load-atoms/datasets/"
BASE_REMOTE_URL = "https://github.com/jla-gardner/load-atoms/raw/main/database/"


def generate_checksum(file_path: Path | str) -> str:
    """Generate a checksum for a file."""

    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()[:12]


def valid_checksum(hash: str) -> bool:
    """Check if a hash is valid."""
    if len(hash) != 12:
        return False
    return all(c in string.hexdigits for c in hash)


def matches_checksum(file_path: Path, hash: str) -> bool:
    """Check if a file matches a given hash."""
    return generate_checksum(file_path) == hash


T = TypeVar("T")
Y = TypeVar("Y")


class LazyMapping(Generic[T, Y]):
    """
    A mapping that lazily loads its values.

    Parameters
    ----------
    keys
        The keys of the mapping.
    loader
        A function that takes a key and returns a value.
    """

    def __init__(
        self,
        keys: Sequence[T],
        loader: Callable[[T], Y],
    ):
        self.keys = keys
        self.loader = loader
        self._mapping = {}

    def __getitem__(self, key: T) -> Y:
        if key not in self.keys:
            raise KeyError(key)
        if key not in self._mapping:
            self._mapping[key] = self.loader(key)
        return self._mapping[key]

    def __contains__(self, key: T):
        return key in self.keys

    def __repr__(self) -> str:
        return f"LazyMapping(keys={self.keys})"


def frontend_url(dataset_info):
    """Get the URL for a dataset's information page."""
    return FRONTEND_URL + dataset_info.name + ".html"


class UnknownDatasetException(Exception):
    def __init__(self, dataset_id):
        super().__init__(f"Unknown dataset: {dataset_id}")


def union(things: Iterable[Iterable]):
    """Get the set union of a list of iterables."""
    return set.union(*map(set, things))


def intersect(things: Iterable[Iterable]):
    """Get the set intersection of a list of iterables."""
    return set.intersection(*map(set, things))


def lpad(s, indent=4):
    return "\n".join(" " * indent + line for line in s.splitlines())



import pytest
from load_atoms.backend import download, download_all

RAW_GITHUB_URL = (
    "https://raw.githubusercontent.com/jla-gardner/load-atoms/main/"
)


def test_download(tmp_path):
    url = RAW_GITHUB_URL + "README.md"
    save_to = tmp_path / "test"

    # check downloading to explicit file works:
    download(url, save_to)
    assert save_to.exists()

    # check downloading to directory works:
    download(url, tmp_path)
    assert (tmp_path / "README.md").exists()

    fake_url = RAW_GITHUB_URL + "fake-file.txt"
    with pytest.raises(Exception):
        download(fake_url, save_to)


def test_download_all(tmp_path):
    # download_all on nothing should do nothing
    download_all([], tmp_path)

    # download a few files we know exist
    files = ["README.md", "LICENSE", ".gitignore"]
    urls = [RAW_GITHUB_URL + f for f in files]
    download_all(urls, tmp_path)
    assert all((tmp_path / f).exists() for f in files)

    # attempt to download a file that doesn't exist
    with pytest.raises(Exception):
        download_all([RAW_GITHUB_URL + "fake-file.txt", *urls], tmp_path)



from pathlib import Path

import numpy as np
import pytest
from ase import Atoms
from ase.io import read, write
from load_atoms import load_dataset
from load_atoms.dataset import AtomsDataset, summarise_dataset
from load_atoms.utils import UnknownDatasetException

STRUCTURES = [Atoms("H2O"), Atoms("H2O2")]


def _is_water_dataset(ds):
    assert len(ds) == 2
    assert set(ds[0].symbols) == {"H", "O"}


def test_dataset_from_structures():
    ds = load_dataset(STRUCTURES)
    _is_water_dataset(ds)


def test_dataset_writeable_and_readable(tmp_path):
    ds = load_dataset(STRUCTURES)
    write(tmp_path / "test.xyz", ds)  # type: ignore

    ds2 = load_dataset(read(tmp_path / "test.xyz", index=":"))  # type: ignore
    _is_water_dataset(ds2)

    ds3 = load_dataset(tmp_path / "test.xyz")
    _is_water_dataset(ds3)


@pytest.mark.filterwarnings("ignore:Creating a dataset with a single structure")
def test_indexing():
    ds = load_dataset(STRUCTURES)

    structure = ds[0]
    assert isinstance(
        structure, Atoms
    ), "Indexing should return an Atoms object"
    assert (
        structure is STRUCTURES[0]
    ), "Indexing should return the same structure"

    sub_dataset = ds[1:]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Slicing should return a dataset"
    assert (
        len(sub_dataset) == 1
    ), "Slicing should return the correct number of structures"

    indices = np.array([0, 1, 0, 1])
    sub_dataset = ds[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 4
    ), "Indexing should return the correct number of structures"

    indices = np.array([True, False])
    sub_dataset = ds[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 1
    ), "Indexing should return the correct number of structures"

    indices = [0, 1, 1]
    sub_dataset = ds[indices]
    assert isinstance(
        sub_dataset, AtomsDataset
    ), "Indexing should return a dataset"
    assert (
        len(sub_dataset) == 3
    ), "Indexing should return the correct number of structures"


def test_can_load_from_id():
    # don't pass a root to mimic the default behaviour
    structures = load_dataset("C-GAP-17")
    assert len(structures) == 4530

    with pytest.raises(UnknownDatasetException):
        load_dataset("made_up_dataset")


def test_summarise():
    ds = load_dataset(STRUCTURES)
    summary = summarise_dataset(ds)
    assert "Dataset" in summary, "The summary should contain the dataset name"

    ds = load_dataset("C-GAP-17")
    summary = repr(ds)
    assert "energy" in summary, "The summary should contain the property names"


def test_useful_error_message():
    with pytest.raises(
        TypeError,
        match="provide a string, a list of structures, or a path to a file.",
    ):
        load_dataset(1)  # type: ignore

    with pytest.raises(
        ValueError,
        match="The provided path does not exist.",
    ):
        load_dataset(Path("made_up_file.xyz"))


def test_useful_warning(tmp_path):
    structure = Atoms("H2O")
    write(tmp_path / "test.xyz", structure)  # type: ignore

    with pytest.warns(
        UserWarning,
        match="single structure",
    ):
        load_dataset(tmp_path / "test.xyz")


def tets_info_and_arrays():
    ds = load_dataset("QM7")

    assert "energy" in ds.info
    assert isinstance(ds.info["energy"], np.ndarray)

    assert "positions" in ds.arrays
    assert ds.arrays["positions"].shape[-1] == 3
